
exports.Lru = Lru;

var hasOwn = Object.prototype.hasOwnProperty;
function Lru(maxLength) {
    if (!maxLength)
        throw new Error("LRU cache must be constructed with a maximum length.");
    var map = {};
    var length = 0;

    var head = {};
    head.next = head;
    head.prev = head;
    function remove(node) {
        delete map[node.key];
        node.prev.next = node.next;
        node.next.prev = node.prev;
        length--;
    }
    function insert(node) {
        map[node.key] = node;
        var prev = head.prev;
        head.prev = node;
        node.prev = prev;
        prev.next = node;
        node.next = head;
        length++;
        if (length > maxLength)
            remove(head.next);
    }

    function get(key) {
        if (!hasOwn.call(map, key))
            throw new ValueError("LRU cache does not contain that key.");
        var node = map[key];
        remove(node);
        insert(node);
        return node.value;
    }
    function set(key, value) {
        var node;
        if (map[key]) {
            node = map[key];
            node.value = value;
            remove(node);
        } else {
            node = {};
            node.key = key;
            node.value = value;
        }
        insert(node);
    }
    function del(key) {
        var node = map[key];
        remove(node);
    }
    function has(key) {
        return hasOwn.call(map, key);
    }
    function keys() {
        return Object.keys(map);
    }

    function toSource() {
        return '[LRU ' + length + ' ' +
            Object.keys(map).map(function (key) {
                var node = map[key];
                return (
                    (node.prev.key || '@') +
                    '<-' + key + ':' + node.value + '->' +
                    (node.next.key || '@')
                );
            }).join(' ') + ']';
    }

    return {
        "get": get,
        "set": set,
        "del": del,
        "has": has,
        "toSource": toSource,
        "toString": toString,
        "keys": keys
    }
}

function demo() {
    var lru = Lru(3);
    console.log(lru.keys());
    lru.set('a', {});
    console.log(lru.keys());
    lru.set('b', {});
    console.log(lru.keys());
    lru.set('c', {});
    console.log(lru.keys());
    lru.get('a');
    console.log(lru.keys());
    lru.set('d', {});
    console.log(lru.keys());
}

if (module === require.main)
    demo()

